# Copyright 2023-2024 The Wazo Authors  (see the AUTHORS file)
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
import unittest
from collections.abc import Iterable
from copy import deepcopy
from typing import Optional
from unittest.mock import Mock, patch
from unittest.mock import sentinel as s

import jwt as jwt_module
import requests
from hamcrest import assert_that, calling, has_properties, not_, raises

sys.modules['wazo_auth'] = Mock()
sys.modules['wazo_auth.http'] = Mock(required_acl=Mock(return_value=lambda x: x))
sys.modules['wazo_auth.exceptions'] = Mock(UnknownExternalAuthConfigException=Exception)
sys.modules['wazo_auth.flask_helpers'] = Mock(
    Tenant=Mock(autodetect=Mock(return_value=Mock(uuid=s.tenant_uuid)))
)
sys.modules['wazo_auth.services.external_auth'] = Mock(ExternalAuthService=Mock())
sys.modules['wazo_auth.plugins.external_auth.mobile.http'] = Mock(
    MobileAuthSenderID=Mock(__name__='MobileAuthSenderID', get=s.MobileAuthSenderID)
)
sys.modules['xivo.rest_api_helpers'] = Mock(APIException=Exception)

from ..fcm import (  # noqa: E402
    ConfigurationError,
    JWTRefreshError,
    Plugin,
    PushServicesUnavailable,
)

PUSH_PROXY_HOST = s.host
PUSH_PROXY_PORT = s.port


class TestConfiguration(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.plugin = Plugin()
        cls.valid_dependencies = {
            'config': {
                'nestbox': {
                    'auth': {
                        'username': s.username,
                        'password': s.password,
                    },
                },
                'mobile_push_proxy_host': PUSH_PROXY_HOST,
                'mobile_push_proxy_port': PUSH_PROXY_PORT,
            },
            'external_auth_service': Mock(),
        }

    @classmethod
    def make_dependencies(cls, remove: Optional[Iterable[str]] = None) -> dict:
        def walk(dict_: dict) -> dict:
            for key, value in tuple(dict_.items()):
                if key in remove:
                    dict_.pop(key)
                elif isinstance(value, dict):
                    walk(value)
            return dependencies

        dependencies = deepcopy(cls.valid_dependencies)
        return walk(dependencies)

    def test_dependencies_missing_parameters(self):
        configurations = [
            (
                self.make_dependencies(remove=['external_auth_service']),
                'Missing dependencies: external_auth_service',
            ),
            (
                self.make_dependencies(remove=['config']),
                'Missing dependencies: config',
            ),
            (
                self.make_dependencies(remove=['mobile_push_proxy_host']),
                'Missing keys in configuration: mobile_push_proxy_host',
            ),
            (
                self.make_dependencies(remove=['mobile_push_proxy_port']),
                'Missing keys in configuration: mobile_push_proxy_port',
            ),
            (
                self.make_dependencies(remove=['nestbox']),
                'Missing keys in configuration: nestbox',
            ),
            (
                self.make_dependencies(remove=['auth']),
                'no authentication section found',
            ),
            (
                self.make_dependencies(remove=['username']),
                'no username found',
            ),
            (
                self.make_dependencies(remove=['password']),
                'no password found',
            ),
        ]

        for dependencies, err_message in configurations:
            assert_that(
                calling(self.plugin._configure).with_args(dependencies),
                raises(ConfigurationError, pattern=err_message),
            )

    def test_valid_dependencies(self):
        assert_that(
            calling(self.plugin.load).with_args(self.valid_dependencies),
            not_(raises(ConfigurationError)),
        )


class TestRoutePatching(unittest.TestCase):
    MODULE_MOBILE_AUTH_SENDER_ID = 'wazo_nestbox_plugin.auth.fcm.MobileAuthSenderID'

    @classmethod
    def setUpClass(cls):
        cls.dependencies = dict(
            config={
                'nestbox': {
                    'auth': {'service_id': s.service_id, 'service_key': s.service_key}
                },
                'mobile_push_proxy_host': PUSH_PROXY_HOST,
                'mobile_push_proxy_port': PUSH_PROXY_PORT,
            },
            external_auth_service=Mock(),
        )
        cls.plugin = Plugin()

    @patch(MODULE_MOBILE_AUTH_SENDER_ID, Mock(__name__=s.name, get=s.get))
    def test_patch_get(self):
        from ..fcm import MobileAuthSenderID

        assert MobileAuthSenderID.get is s.get

        self.plugin.load(self.dependencies)

        assert MobileAuthSenderID.get is not s.get


class TestJWT(unittest.TestCase):
    MODULE_JWT_DECODE = 'wazo_nestbox_plugin.auth.fcm.jwt_decode'
    MODULE_AUTH_CLIENT = 'wazo_nestbox_plugin.auth.fcm.AuthClient'

    @classmethod
    def setUpClass(cls):
        dependencies = dict(
            config={
                'nestbox': {
                    'auth': {'service_id': s.service_id, 'service_key': s.service_key}
                },
                'mobile_push_proxy_host': PUSH_PROXY_HOST,
                'mobile_push_proxy_port': PUSH_PROXY_PORT,
            },
            external_auth_service=Mock(),
        )
        cls.plugin = Plugin()
        cls.plugin.load(dependencies)

    def setUp(self):
        self.auth_client = Mock(return_value=Mock(token=Mock(new=Mock())))
        AuthClientMock = Mock(return_value=self.auth_client)
        patch(self.MODULE_AUTH_CLIENT, AuthClientMock).start()

    def tearDown(self):
        patch.stopall()

    def test_refresh_jwt_when_jwt_is_valid(self):
        self.auth_client.token.new.return_value = {'metadata': {'jwt': 'new jwt'}}
        self.plugin._jwt = 'the jwt'

        with patch(self.MODULE_JWT_DECODE, Mock(return_value=True)):
            assert self.plugin._refresh_jwt() == 'the jwt'

    def test_refresh_jwt_when_stored_jwt_is_invalid(self):
        self.auth_client.token.new.return_value = {'metadata': {'jwt': 'new jwt'}}

        exceptions = (
            jwt_module.ExpiredSignatureError,
            jwt_module.DecodeError,
            AttributeError,
        )

        for exception in exceptions:
            self.plugin._jwt = 'the jwt'

            with patch(self.MODULE_JWT_DECODE, Mock(side_effect=exception)):
                assert self.plugin._refresh_jwt() == 'new jwt'

    def test_refresh_jwt_when_network_error(self):
        self.auth_client.token.new.side_effect = requests.HTTPError
        self.plugin._jwt = 'the jwt'

        assert_that(calling(self.plugin._refresh_jwt), raises(JWTRefreshError))

    def test_refresh_jwt_when_token_has_no_jwt(self):
        self.auth_client.token.new.return_value = {'metadata': {}}
        self.plugin._jwt = 'the jwt'

        assert_that(calling(self.plugin._refresh_jwt), raises(JWTRefreshError))


class TestFetchSenderID(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dependencies = dict(
            config={
                'nestbox': {
                    'auth': {'service_id': s.service_id, 'service_key': s.service_key}
                },
                'mobile_push_proxy_host': PUSH_PROXY_HOST,
                'mobile_push_proxy_port': PUSH_PROXY_PORT,
            },
            external_auth_service=Mock(),
        )
        cls.plugin = Plugin()
        cls.plugin.load(cls.dependencies)
        cls.expected_url = f'https://{PUSH_PROXY_HOST}:{PUSH_PROXY_PORT}/fcm/sender-id'
        cls.expected_headers = dict(Authorization=f'key={s.jwt}')

    def setUp(self):
        self.plugin._refresh_jwt = Mock(return_value=s.jwt)
        self.response = Mock(
            raise_for_status=Mock(), json=Mock(return_value={'sender_id': s.sender_id})
        )
        self.requests_get = Mock(return_value=self.response)
        patch('wazo_nestbox_plugin.auth.fcm.requests.get', self.requests_get).start()
        self.plugin._fetch_sender_id.cache_clear()

    def tearDown(self):
        patch.stopall()

    def test_fetch_sender_id(self):
        result = self.plugin._fetch_sender_id()

        self.requests_get.assert_called_once_with(
            self.expected_url, headers={'Authorization': f'key={s.jwt}'}, timeout=10
        )
        assert result is s.sender_id
        assert_that(
            self.plugin._fetch_sender_id.cache_info(),
            has_properties(hits=0, misses=1, currsize=1),
        )

    def test_fetch_sender_id_updates_last_known_sender_id_when_successful(self):
        self.plugin._last_known_sender_id = s.last_sender_id
        self.response.json.return_value = {'sender_id': s.new_sender_id}

        assert self.plugin._last_known_sender_id == s.last_sender_id

        result = self.plugin._fetch_sender_id()

        assert result == s.new_sender_id
        assert self.plugin._last_known_sender_id == s.new_sender_id

    def test_fetch_sender_id_does_not_update_last_known_sender_id_when_failure(self):
        self.plugin._last_known_sender_id = s.last_sender_id
        self.response.raise_for_status.side_effect = requests.HTTPError

        assert self.plugin._last_known_sender_id == s.last_sender_id

        assert_that(
            calling(self.plugin._fetch_sender_id), raises(PushServicesUnavailable)
        )
        assert self.plugin._last_known_sender_id == s.last_sender_id

    def test_fetch_sender_id_caching(self):
        for _ in range(5):
            self.plugin._fetch_sender_id()

        self.requests_get.assert_called_once_with(
            self.expected_url, headers=self.expected_headers, timeout=10
        )
        assert_that(
            self.plugin._fetch_sender_id.cache_info(),
            has_properties(hits=4, misses=1, currsize=1),
        )

    def test_fetch_sender_id_when_network_error(self):
        self.response.raise_for_status.side_effect = requests.HTTPError

        assert_that(
            calling(self.plugin._fetch_sender_id), raises(PushServicesUnavailable)
        )
        assert_that(
            self.plugin._fetch_sender_id.cache_info(),
            has_properties(hits=0, misses=1, currsize=0),
        )

    def test_fetch_sender_id_when_jwt_error(self):
        self.plugin._refresh_jwt.side_effect = JWTRefreshError()

        assert_that(calling(self.plugin._fetch_sender_id), raises(JWTRefreshError))
        assert_that(
            self.plugin._fetch_sender_id.cache_info(),
            has_properties(hits=0, misses=1, currsize=0),
        )


class TestGET(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dependencies = dict(
            config={
                'nestbox': {
                    'auth': {'service_id': s.service_id, 'service_key': s.service_key}
                },
                'mobile_push_proxy_host': PUSH_PROXY_HOST,
                'mobile_push_proxy_port': PUSH_PROXY_PORT,
            },
            external_auth_service=Mock(),
        )
        cls.plugin = Plugin()
        cls.mobile_auth_sender_id = Mock()
        cls.plugin.load(cls.dependencies)

    def setUp(self):
        self.plugin._fetch_sender_id = Mock(return_value=s.sender_id)
        self.plugin._sender_id_from_mobile_config = Mock(return_value=None)

    def test_get_when_portal_is_down_with_no_last_known_sender_id(self):
        self.plugin._last_known_sender_id = None
        self.plugin._fetch_sender_id.side_effect = JWTRefreshError

        assert_that(
            calling(self.plugin._patched_get).with_args(Mock(), s.user_uuid),
            raises(JWTRefreshError),
        )

    def test_get_sender_id_when_portal_is_down(self):
        self.plugin._last_known_sender_id = s.last_sender_id
        self.plugin._fetch_sender_id.side_effect = JWTRefreshError

        response, status_code = self.plugin._patched_get(Mock(), s.user_uuid)

        assert status_code == 200
        assert response['sender_id'] == s.last_sender_id

    def test_get_when_push_proxy_is_down_with_no_known_sender_id(self):
        self.plugin._last_known_sender_id = None
        self.plugin._fetch_sender_id.side_effect = PushServicesUnavailable

        assert_that(
            calling(self.plugin._patched_get).with_args(Mock(), s.user_uuid),
            raises(PushServicesUnavailable),
        )

    def test_get_sender_id_when_push_proxy_is_down(self):
        self.plugin._last_known_sender_id = s.last_sender_id
        self.plugin._fetch_sender_id.side_effect = PushServicesUnavailable

        response, status_code = self.plugin._patched_get(Mock(), s.user_uuid)

        assert status_code == 200
        assert response['sender_id'] == s.last_sender_id

    def test_prioritize_sender_id_from_mobile_config_when_available(self):
        self.plugin._last_known_sender_id = s.last_sender_id
        self.plugin._sender_id_from_mobile_config.return_value = s.stack_sender_id
        self.plugin._fetch_sender_id.return_value = s.new_sender_id

        response, status_code = self.plugin._patched_get(Mock(), s.user_uuid)

        assert status_code == 200
        assert response['sender_id'] == s.stack_sender_id
